from collections import abc
from pandas._typing import ArrayLike, FilePathOrBuffer as FilePathOrBuffer, Scalar
from pandas.core.frame import DataFrame as DataFrame
from typing import Any, Callable, Dict, List, Literal, Optional, AnyStr, overload, Sequence, Union

# not checking types in Callable params for read_csv, the documentation is too ambiguous
AnyCallable = Callable[[Any], Any]

@overload
def read_csv(filepath_or_buffer: Any, sep: str = ..., delimiter: Optional[str] = ..., header: Optional[Union[int, List[int], str]] = ...,
             names: Optional[ArrayLike] = ..., index_col: Optional[Union[int, str, Sequence[int], Sequence[str], bool]] = None,
             usecols: Optional[Union[ArrayLike, AnyCallable]] = ..., squeeze: bool = ..., prefix: str = ..., mangle_dupe_cols: bool = ...,
             dtype: Any = ..., engine: Optional[Literal['c', 'python']] = ..., converters: Optional[Dict[Any, Any]] = ...,
             true_values: Optional[List[Any]] = ..., false_values: Optional[List[Any]] = ..., skipinitialspace: bool = ...,
             skiprows: Optional[Union[ArrayLike, int, AnyCallable]] = ..., skipfooter: int = ..., nrows: Optional[int] = ...,
             na_values: Optional[Union[Scalar, str, ArrayLike, Dict[Any, Any]]] = ..., keep_default_na: bool = ...,
             na_filter: bool = ..., verbose: bool = ..., skip_blank_lines: bool = ...,
             parse_dates: Union[bool, List[int], List[str], List[List[Any]], Dict[Any, Any]] = ..., infer_datetime_format: bool = ...,
             keep_date_col: bool = ..., date_parser: Optional[AnyCallable] = ..., dayfirst: bool = ..., cache_dates: bool = ...,
             *, iterator: Literal[True], chunksize: Optional[int] = ..., **kwargs: Any) -> TextFileReader: ...

@overload
def read_csv(filepath_or_buffer: Any, sep: str = ..., delimiter: Optional[str] = ..., header: Optional[Union[int, List[int], str]] = ...,
             names: Optional[ArrayLike] = ..., index_col: Optional[Union[int, str, Sequence[int], Sequence[str], bool]] = None,
             usecols: Optional[Union[ArrayLike, AnyCallable]] = ..., squeeze: bool = ..., prefix: str = ..., mangle_dupe_cols: bool = ...,
             dtype: Any = ..., engine: Optional[Literal['c', 'python']] = ..., converters: Optional[Dict[Any, Any]] = ...,
             true_values: Optional[List[Any]] = ..., false_values: Optional[List[Any]] = ..., skipinitialspace: bool = ...,
             skiprows: Optional[Union[ArrayLike, int, AnyCallable]] = ..., skipfooter: int = ..., nrows: Optional[int] = ...,
             na_values: Optional[Union[Scalar, str, ArrayLike, Dict[Any, Any]]] = ..., keep_default_na: bool = ...,
             na_filter: bool = ..., verbose: bool = ..., skip_blank_lines: bool = ...,
             parse_dates: Union[bool, List[int], List[str], List[List[Any]], Dict[Any, Any]] = ..., infer_datetime_format: bool = ...,
             keep_date_col: bool = ..., date_parser: Optional[AnyCallable] = ..., dayfirst: bool = ..., cache_dates: bool = ...,
             *, iterator: Literal[False] = ..., chunksize: int, **kwargs: Any) -> TextFileReader: ...

@overload
def read_csv(filepath_or_buffer: Any, sep: str = ..., delimiter: Optional[str] = ..., header: Optional[Union[int, List[int], str]] = ...,
             names: Optional[ArrayLike] = ..., index_col: Optional[Union[int, str, Sequence[int], Sequence[str], bool]] = None,
             usecols: Optional[Union[ArrayLike, AnyCallable]] = ..., squeeze: bool = ..., prefix: str = ..., mangle_dupe_cols: bool = ...,
             dtype: Any = ..., engine: Optional[Literal['c', 'python']] = ..., converters: Optional[Dict[Any, Any]] = ...,
             true_values: Optional[List[Any]] = ..., false_values: Optional[List[Any]] = ..., skipinitialspace: bool = ...,
             skiprows: Optional[Union[ArrayLike, int, AnyCallable]] = ..., skipfooter: int = ..., nrows: Optional[int] = ...,
             na_values: Optional[Union[Scalar, str, ArrayLike, Dict[Any, Any]]] = ..., keep_default_na: bool = ...,
             na_filter: bool = ..., verbose: bool = ..., skip_blank_lines: bool = ...,
             parse_dates: Union[bool, List[int], List[str], List[List[Any]], Dict[Any, Any]] = ..., infer_datetime_format: bool = ...,
             keep_date_col: bool = ..., date_parser: Optional[AnyCallable] = ..., dayfirst: bool = ..., cache_dates: bool = ...,
             *, iterator: Literal[False] = ..., chunksize: None = ..., **kwargs: Any) -> DataFrame: ...

read_table: Any

def read_fwf(filepath_or_buffer: FilePathOrBuffer[AnyStr], colspecs: Any = ..., widths: Any = ..., infer_nrows: Any = ..., **kwds: Any) -> Any: ...

class TextFileReader(abc.Iterator[Any]):
    f: Any = ...
    orig_options: Any = ...
    engine: Any = ...
    chunksize: Any = ...
    nrows: Any = ...
    squeeze: Any = ...
    def __init__(self, f: Any, engine: Optional[Any] = ..., **kwds: Any) -> None: ...
    def close(self) -> None: ...
    def __next__(self) -> Any: ...
    def read(self, nrows: Optional[Any] = ...) -> Any: ...
    def get_chunk(self, size: Optional[Any] = ...) -> Any: ...

def TextParser(*args: Any, **kwds: Any) -> Any: ...


import sys
from datetime import timedelta

from pandas._libs.tslibs.timedeltas import Timedelta as Timedelta, TimedeltaUnitChoice
from pandas.core.dtypes.generic import ABCIndexClass as ABCIndexClass, ABCSeries as ABCSeries
from typing import overload, List, Union

from pandas.core.indexes.timedeltas import TimedeltaIndex
from pandas.core.series import Series

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

ErrorOption = Literal["ignore", "raise", "coerce"]

@overload
def to_timedelta(arg: timedelta, errors: ErrorOption = ...) -> Timedelta: ...
@overload
def to_timedelta(arg: Union[float, int], unit: TimedeltaUnitChoice = ..., errors: ErrorOption = ...) -> Timedelta: ...
@overload
def to_timedelta(arg: List[Union[float, int]], unit: TimedeltaUnitChoice = ..., errors: ErrorOption = ...) -> TimedeltaIndex: ...
@overload
def to_timedelta(arg: str, errors: ErrorOption = ...) -> Timedelta: ...
@overload
def to_timedelta(arg: List[Union[str]], errors: ErrorOption = ...) -> TimedeltaIndex: ...
@overload
def to_timedelta(arg: Series, unit: TimedeltaUnitChoice = ..., errors: ErrorOption = ...) -> Series: ...



import sys
from datetime import timedelta
from typing import Union, overload

import numpy as np

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

TimeValue = Union[float, int]

TimedeltaUnitsNano = Literal["ns", "nanoseconds", "nano", "nanos", "nanosecond", "n"]
TimedeltaUnitsMicro = Literal["us","microseconds", "microsecond", "µs", "micro", "micros", "u"]
TimedeltaUnitsMillis = Literal["ms", "milliseconds", "millisecond", "milli", "millis", "l"]
TimedeltaUnitsSec = Literal["s", "seconds", "sec", "second"]
TimedeltaUnitsCalendar = Literal["W", "w", "D", "d", "days", "day"]
TimedeltaUnitsClock = Literal["hours", "hour", "hr", "h", "m", "minute", "min", "minutes", "t"]

TimedeltaUnitChoice = Literal[TimedeltaUnitsClock, TimedeltaUnitsCalendar, TimedeltaUnitsSec,
                              TimedeltaUnitsMicro, TimedeltaUnitsMillis, TimedeltaUnitsNano]

class Timedelta(timedelta):

    @overload
    def __init__(self, value: Union[Timedelta, timedelta, np.timedelta64, float, int], unit: TimedeltaUnitChoice = ...) -> None: ...
    @overload
    def __init__(self, value: str) -> None: ...
    # Wee need so many overloads to ensue that at least one keyword argument is passed e.g. pd.Timedelta() is invalid
    @overload
    def __init__(self, weeks: TimeValue, days: TimeValue = ..., hours: TimeValue = ..., minutes: TimeValue = ..., seconds: TimeValue = ..., milliseconds: TimeValue = ..., microseconds: TimeValue = ..., nanoseconds: TimeValue = ...) -> None: ...
    @overload
    def __init__(self, days: TimeValue, weeks: TimeValue = ..., hours: TimeValue = ..., minutes: TimeValue = ..., seconds: TimeValue = ..., milliseconds: TimeValue = ..., microseconds: TimeValue = ..., nanoseconds: TimeValue = ...) -> None: ...
    @overload
    def __init__(self, hours: TimeValue, days: TimeValue = ..., weeks: TimeValue = ..., minutes: TimeValue = ..., seconds: TimeValue = ..., milliseconds: TimeValue = ..., microseconds: TimeValue = ..., nanoseconds: TimeValue = ...) -> None: ...
    @overload
    def __init__(self, minutes: TimeValue, days: TimeValue = ..., hours: TimeValue = ..., weeks: TimeValue = ..., seconds: TimeValue = ..., milliseconds: TimeValue = ..., microseconds: TimeValue = ..., nanoseconds: TimeValue = ...) -> None: ...
    @overload
    def __init__(self, seconds: TimeValue, days: TimeValue = ..., hours: TimeValue = ..., minutes: TimeValue = ..., weeks: TimeValue = ..., milliseconds: TimeValue = ..., microseconds: TimeValue = ..., nanoseconds: TimeValue = ...) -> None: ...
    @overload
    def __init__(self, milliseconds: TimeValue, days: TimeValue = ..., hours: TimeValue = ..., minutes: TimeValue = ..., seconds: TimeValue = ..., weeks: TimeValue = ..., microseconds: TimeValue = ..., nanoseconds: TimeValue = ...) -> None: ...
    @overload
    def __init__(self, microseconds: TimeValue, days: TimeValue = ..., hours: TimeValue = ..., minutes: TimeValue = ..., seconds: TimeValue = ..., milliseconds: TimeValue = ..., weeks: TimeValue = ..., nanoseconds: TimeValue = ...) -> None: ...
    @overload
    def __init__(self, nanoseconds: TimeValue, days: TimeValue = ..., hours: TimeValue = ..., minutes: TimeValue = ..., seconds: TimeValue = ..., milliseconds: TimeValue = ..., microseconds: TimeValue = ..., weeks: TimeValue = ...) -> None: ...



from pandas.io.parsers.readers import TextFileReader as TextFileReader
from pandas.io.parsers.readers import TextParser as TextParser
from pandas.io.parsers.readers import read_csv as read_csv
from pandas.io.parsers.readers import read_fwf as read_fwf
from pandas.io.parsers.readers import read_table as read_table


from typing import Any, Optional

from pandas.io.parsers.base_parser import ParserBase


class CParserWrapper(ParserBase):
    kwds: Any = ...
    unnamed_cols: Any = ...
    names: Any = ...
    orig_names: Any = ...
    index_names: Any = ...
    def __init__(self, src: Any, **kwds: Any) -> None: ...
    def close(self) -> None: ...
    def set_error_bad_lines(self, status: Any) -> None: ...
    def read(self, nrows: Optional[Any] = ...) -> Any: ...


from collections import abc
from typing import Any, Optional

from pandas.io.parsers.base_parser import ParserBase


class PythonParser(ParserBase):
    data: Any = ...
    buf: Any = ...
    pos: int = ...
    line_pos: int = ...
    encoding: Any = ...
    compression: Any = ...
    memory_map: Any = ...
    skiprows: Any = ...
    skipfunc: Any = ...
    skipfooter: Any = ...
    delimiter: Any = ...
    quotechar: Any = ...
    escapechar: Any = ...
    doublequote: Any = ...
    skipinitialspace: Any = ...
    lineterminator: Any = ...
    quoting: Any = ...
    skip_blank_lines: Any = ...
    warn_bad_lines: Any = ...
    error_bad_lines: Any = ...
    names_passed: Any = ...
    has_index_names: bool = ...
    verbose: Any = ...
    converters: Any = ...
    dtype: Any = ...
    thousands: Any = ...
    decimal: Any = ...
    comment: Any = ...
    num_original_columns: Any = ...
    columns: Any = ...
    orig_names: Any = ...
    index_names: Any = ...
    nonnum: Any = ...
    def __init__(self, f: Any, **kwds: Any) -> None: ...
    def read(self, rows: Optional[Any] = ...) -> Any: ...
    def get_chunk(self, size: Optional[Any] = ...) -> Any: ...

class FixedWidthReader(abc.Iterator[Any]):
    f: Any = ...
    buffer: Any = ...
    delimiter: Any = ...
    comment: Any = ...
    colspecs: Any = ...
    def __init__(self, f: Any, colspecs: Any, delimiter: Any, comment: Any, skiprows: Optional[Any] = ..., infer_nrows: int = ...) -> None: ...
    def get_rows(self, infer_nrows: Any, skiprows: Optional[Any] = ...) -> Any: ...
    def detect_colspecs(self, infer_nrows: int = ..., skiprows: Optional[Any] = ...) -> Any: ...
    def __next__(self) -> Any: ...

class FixedWidthFieldParser(PythonParser):
    colspecs: Any = ...
    infer_nrows: Any = ...
    def __init__(self, f: Any, **kwds: Any) -> None: ...

def count_empty_vals(vals: Any) -> Any: ...


import datetime
import sys
from typing import Any, Union, Tuple

import numpy as np

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

ClosedOption = Literal['left', 'right', 'both', 'neither']
# According to docs "ValueError: Only numeric, Timestamp and Timedelta endpoints are allowed when constructing an Interval."
IntervalBound = Union[int, float, np.integer, np.complex_, datetime.datetime, datetime.timedelta]

class IntervalMixin:
    @property
    def closed_left(self) -> bool: ...
    @property
    def closed_right(self) -> bool: ...
    @property
    def open_left(self) -> bool: ...
    @property
    def open_right(self) -> bool: ...
    @property
    def mid(self) -> Any: ...
    @property
    def length(self) -> Any: ...
    @property
    def is_empty(self) -> bool: ...

class Interval(IntervalMixin):
    @property
    def closed(self) -> str: ...
    @property
    def left(self) -> IntervalBound: ...
    @property
    def right(self) -> IntervalBound: ...
    def __init__(self, left: IntervalBound, right: IntervalBound, closed: ClosedOption = ...) -> None: ...
    def __hash__(self) -> int: ...
    def __contains__(self, key: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...

    def __add__(self, y: Any) -> Interval: ...
    def __sub__(self, y: Any) -> Interval: ...
    def __mul__(self, y: Any) -> Interval: ...
    def __truediv__(self, y: Any) -> Interval: ...
    def __floordiv__(self, y: Any) -> Interval: ...
    def overlaps(self, other: Interval) -> bool: ...


def intervals_to_interval_bounds(intervals: np.ndarray, validate_closed: bool = ...) -> Tuple[np.ndarray, np.ndarray, str]: ...


class IntervalTree: ...


from typing import Any


class ParserBase:
    names: Any = ...
    orig_names: Any = ...
    prefix: Any = ...
    index_col: Any = ...
    unnamed_cols: Any = ...
    index_names: Any = ...
    col_names: Any = ...
    parse_dates: Any = ...
    date_parser: Any = ...
    dayfirst: Any = ...
    keep_date_col: Any = ...
    na_values: Any = ...
    na_fvalues: Any = ...
    na_filter: Any = ...
    keep_default_na: Any = ...
    true_values: Any = ...
    false_values: Any = ...
    mangle_dupe_cols: Any = ...
    infer_datetime_format: Any = ...
    cache_dates: Any = ...
    header: Any = ...
    handles: Any = ...
    def __init__(self, kwds: Any) -> None: ...
    def close(self) -> None: ...


from typing import Any, Mapping, overload, Sequence, Union

from pandas import DataFrame as DataFrame, Series as Series
from pandas._typing import AxisOptionHorizontal, AxisOptionVertical, FrameOrSeries, Label


# For some reason mypy won't read this declaration correctly if we use 2 overloads with union
# It's complicated to read so citing the docs:
# "When concatenating all Series along the index (axis=0), a Series is returned. When objs contains at least one DataFrame,
# a DataFrame is returned. When concatenating along the columns (axis=1), a DataFrame is returned."
# Label=Optional[Hashable] should be used instead of Any as key objs Mapping however mypy isn't handling it correctly
@overload
def concat(objs: Union[Sequence[Series], Mapping[Any, Series]], axis: AxisOptionHorizontal = ..., join: str=..., ignore_index: bool=..., keys: Any = ..., levels: Any = ..., names: Any = ..., verify_integrity: bool=..., sort: bool=..., copy: bool=...) -> Series: ...
@overload
def concat(objs: Union[Sequence[DataFrame], Mapping[Any, DataFrame]], axis: AxisOptionHorizontal = ..., join: str=..., ignore_index: bool=..., keys: Any = ..., levels: Any = ..., names: Any = ..., verify_integrity: bool=..., sort: bool=..., copy: bool=...) -> DataFrame: ...
@overload
def concat(objs: Union[Sequence[FrameOrSeries], Mapping[Any, FrameOrSeries]], axis: AxisOptionVertical, join: str=..., ignore_index: bool=..., keys: Any = ..., levels: Any = ..., names: Any = ..., verify_integrity: bool=..., sort: bool=..., copy: bool=...) -> DataFrame: ...

class _Concatenator:
    intersect: bool = ...
    objs: Any = ...
    axis: Any = ...
    keys: Any = ...
    names: Any = ...
    levels: Any = ...
    sort: Any = ...
    ignore_index: Any = ...
    verify_integrity: Any = ...
    copy: Any = ...
    new_axes: Any = ...
    def __init__(self, objs: Any, axis: Any = ..., join: str=..., keys: Any = ..., levels: Any = ..., names: Any = ..., ignore_index: bool=..., verify_integrity: bool=..., copy: bool=..., sort: Any = ...) -> None: ...
    def get_result(self) -> Any: ...
